# frozen_string_literal: true

require 'rack'
require 'json'
require 'uri'

module Flon
  # Represents a request.
  class Request
    # @return [String] the body of the request
    attr_reader :body

    # @return [Hash{String => String}] the headers of the request
    attr_reader :headers

    # @return [Symbol] the request's method as a lowercase symbol
    attr_reader :method

    # @return [Hash{String => String}] the merged hash of {route_params} and
    #   {query_params}. Route parameters take precedence
    attr_reader :params

    # @return [String] the request's path
    attr_reader :path

    # @return [Hash{Symbol => String}] the query parameters from the query
    #   string
    attr_reader :query_params

    # @return [Hash{Symbol => String}] the route parameters
    attr_reader :route_params

    # @return [URI] the URL associated with this request
    attr_reader :url

    # @api private
    def initialize(method, rack_request, route_params, body)
      @body = body
      @headers = rack_request.env
      @method = method
      @query_params = rack_request.params
      @route_params = route_params
      @params = coalesce_params(@query_params, @route_params)
      @path = rack_request.path_info
      @url = make_url(rack_request)
    end

    private

    def coalesce_params(query_params, route_params)
      query_params.transform_keys(&:to_sym).merge(route_params)
    end

    def make_url(request)
      URI.parse(request.fullpath)
    end
  end

  # Represents a response, with a status, body, and headers.
  #
  # @!attribute [r] status
  #   @return [Integer] the status to use when responding
  #
  # @!attribute [r] body
  #   @return [#to_json] the object to respond with that shall be #to_json'd
  #
  # @!attribute [r] headers
  #   @return [Hash{String => #to_s}] the headers hash; each value will be
  #     converted with a String with #to_s
  Response = Struct.new(:status, :body, :headers)

  # This class is Flon's Rack application.
  class API
    DEFAULT_RESPONSE_HEADERS = { 'Content-Type' => 'application/json' }.freeze
    private_constant :DEFAULT_RESPONSE_HEADERS

    # Creates a new {API} object with the given API.
    #
    # @param [Object] api the API to use; its class must respond to #router and
    #   the result must be a {Router}.
    # @return [API] the newly constructed object
    def initialize(api)
      raise ArgumentError, "given API's class does not respond to #routes" unless valid_api?(api)

      @api    = api
      @router = api.class.router
    end

    # Implements the Rack interface.
    #
    # @param [Hash] env a rack environment
    # @return [Array] a rack-suitable response
    def call(env)
      request = Rack::Request.new(env)

      method = http_method_to_symbol(request.request_method)
      path   = request.path_info

      response = case (match = @router.match(method, path))
                 when :bad_path   then Response.new(404, '"404 Not Found"')
                 when :bad_method then Response.new(405, '"405 Method Not Allowed"')
                 else dispatch(method, request, match)
                 end

      rack_response(method, response)
    end

    private

    def dispatch(method, rack_request, match)
      if receives_body?(method)
        body = normalise_body(rack_request)
        return Response.new(415, '"415 Unsupported Media Type"') if body == :fail
      else
        body = nil
      end

      request = Request.new(method, rack_request, match.params, body)
      call_action(match.action, request)
    end

    def normalise_body(request)
      body = request.body.read
      return body if body.empty?

      return :fail unless request.get_header('Content-Type') == 'application/json'

      begin
        JSON.parse(body)
      rescue JSON::JSONError
        :fail
      end
    end

    def receives_body?(http_method)
      http_method != :get && http_method != :delete
    end

    def call_action(action, request)
      responsify(call_respect_arity(action.bind(@api), request))
    end

    def responsify(result)
      if result.is_a?(Response)
        result.tap { |it| it.body = it.body.to_json }
      else
        Response.new(200, result.to_json)
      end
    end

    def call_respect_arity(method, *args)
      arity = method.arity
      arity.negative? ? method.call(*args) : method.call(*args[0...arity])
    end

    def rack_response(method, response)
      [
        response.status || 200,
        DEFAULT_RESPONSE_HEADERS.merge(stringify_values(response.headers || {})),
        method == :head || !response.body ? [] : [response.body]
      ]
    end

    def stringify_values(hash)
      hash.transform_values(&:to_s)
    end

    def http_method_to_symbol(http_method)
      http_method.downcase.to_sym
    end

    def valid_api?(api)
      api.class.respond_to?(:router) && api.class.router.is_a?(Flon::Router)
    end
  end
end
