# frozen_string_literal: true

require 'flon/router'

module Flon
  # This module holds Flon's DSL. You should extend it in your class.
  #
  # @example
  #   class BreadAPI
  #     extend Flon::DSL
  #
  #     get '/bread'
  #     def bread
  #       'bread.'
  #     end
  #   end
  module DSL
    # @!group Routing

    # @!method get(path)
    #   Creates a GET route from path, bound to the next defined method.
    #   @param [String] path the path to bind to
    #   @return [void]
    # @!method post(path)
    #   Creates a POST route from path, bound to the next defined method.
    #   @param [String] path the path to bind to
    #   @return [void]
    # @!method put(path)
    #   Creates a PUT route from path, bound to the next defined method.
    #   @param [String] path the path to bind to
    #   @return [void]
    # @!method delete(path)
    #   Creates a DELETE route from path, bound to the next defined method.
    #   @param [String] path the path to bind to
    #   @return [void]
    # @!method patch(path)
    #   Creates a PATCH route from path, bound to the next defined method.
    #   @param [String] path the path to bind to
    #   @return [void]
    %i[get post put delete patch].each do |http_method|
      define_method(http_method) do |path|
        @current_route = [http_method, namespaces.join + path]
      end
    end

    # Creates a namespace of the given path. If a block is given, the namespace
    # is local to the block. If not, it is applied to the rest of the API. You
    # may not call this method without a block twice.
    #
    # @param [String] path the path to use
    # @return [void]
    def namespace(path)
      if block_given?
        namespaces.push(path)
        yield
        namespaces.pop
      else
        unless namespaces.empty?
          raise Flon::Error, 'cannot declare a nested global namespace or declare a global namespace twice'
        end

        namespaces.push(path)
      end
    end

    alias version namespace

    # @!endgroup

    # @api private
    def router
      @router ||= Flon::Router.new
    end

    # @api private
    def namespaces
      @namespaces ||= []
    end

    # @api private
    def method_added(name)
      return unless @current_route

      http_method, path = @current_route
      router.add_route(http_method, path, instance_method(name))

      @current_route = nil
    end
  end
end
